﻿#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using HeuristicLab.Analysis;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Problems.DataAnalysis;
using HEAL.Attic;

namespace HeuristicLab.DatastreamAnalysis {
  internal enum DatastreamAnalysisOptimizerAction {
    None,
    Prepare,
    Start,
    Stop,
    Pause
  }

  [Item("DatastreamAnalysis Optimizer",
     "The main loop for evaluating ensemble models against a incoming datastream of time series fashion.")]
  [Creatable(CreatableAttribute.Categories.Algorithms)]
  [StorableType("C2743DA1-081A-4475-8C2C-BFFEACB44134")]
  public class DatastreamAnalysisOptimizer : Executable, IOptimizer, IStorableContent {
    #region properties
    public string Filename { get; set; }

    private DatastreamAnalysisOptimizerAction daoAction;

    public IEnumerable<IOptimizer> NestedOptimizers { get; }

    private static Color[] _colorsBrightPastel = {
                           Color.FromArgb(65, 140, 240),
                           Color.FromArgb(252, 180, 65),
                           Color.FromArgb(224, 64, 10),
                           Color.FromArgb(5, 100, 146),
                           Color.FromArgb(191, 191, 191),
                           Color.FromArgb(26, 59, 105),
                           Color.FromArgb(255, 227, 130),
                           Color.FromArgb(18, 156, 221),
                           Color.FromArgb(202, 107, 75),
                           Color.FromArgb(0, 92, 219),
                           Color.FromArgb(243, 210, 136),
                           Color.FromArgb(80, 99, 129),
                           Color.FromArgb(241, 185, 168),
                           Color.FromArgb(224, 131, 10),
                           Color.FromArgb(120, 147, 190)
                         };

    [Storable]
    protected ILog log;
    public ILog Log {
      get { return log; }
    }

    [Storable]
    private ResultCollection results;

    public ResultCollection Results {
      get { return results; }
    }

    //private double lastStateValue;
    //private double pR2Sum;
    //private int detectionCount;
    //private int evaluationCount;
    //private double exclusivitySum;
    //private double accuracySum;

    private CancellationTokenSource cancellationTokenSource;
    private bool stopPending;
    private DateTime lastUpdateTime;
    private bool prepared;
    private bool finished;

    [Storable]
    protected int runsCounter;

    [Storable]
    private RunCollection runs;

    public RunCollection Runs
    {
      get { return runs; }
      protected set
      {
        if (value == null) throw new ArgumentNullException();
        if (runs != value) {
          if (runs != null) DeregisterRunsEvents();
          runs = value;
          if (runs != null) RegisterRunsEvents();
        }
      }
    }

    [Storable]
    private IItemList<ProxyEnsembleModel> ensembles;

    public IItemList<ProxyEnsembleModel> Ensembles {
      get { return ensembles; }
      set {
        if (value == null || value == ensembles)
          return;
        if (!(value is IProxyEnsembleModel)) throw new ArgumentException("Invalid ensemble model type");
        DeregisterEnsembleEvents();
        ensembles = value;
        RegisterEnsembleEvents();
        OnEnsemblesChanged();
        Prepare();
      }
    }


    // VAR 1: datastream ~= problem data, VAR 2 (TODO): datastream = external source e.g. webservice, AMQP-Queue, etc.
    [Storable]
    private Datastream datastream;

    public Datastream Datastream {
      get { return datastream; }
      set {
        if (value == null || value == datastream)
          return;
        if(!(value is IDatastream)) throw new ArgumentException("Invalid datastream type");
        DeregisterDatastreamEvents();
        datastream = value;
        RegisterDatastreamEvents();
        OnDatastreamChanged();
        Prepare();
      }
    }

    #endregion properties

    #region results properties
    private int ResultsSlidingWindowMovements {
      get { return ((IntValue)Results["Sliding Window Movements"].Value).Value; }
      set { ((IntValue)Results["Sliding Window Movements"].Value).Value = value; }
    }
    private DataTable ResultsQualities {
      get { return ((DataTable)Results["Qualities 1"].Value); }
    }

    private DataTable ResultsHorizonQualities {
      get { return ((DataTable)Results["Qualities 2"].Value); }
    }

    //private DataTable ResultsPeakQualities {
    //  get { return ((DataTable) Results["Qualities 2"].Value); }
    //}

    private const string ResultsQualitiesMSE = "Mean squared error";
    private const string ResultsQualitiesPR2 = "Pearson R²";

    private ExtendedDataTable ResultsConfidenceTargets {
      get { return ((ExtendedDataTable)Results["ConfidenceTargets"].Value); }
    }

    private ExtendedDataTable ResultsErrorRangeTargets {
      get { return ((ExtendedDataTable)Results["ErrorRangeTargets"].Value); }
    }

    private DataTable ResultsTargets {
      get { return ((DataTable) Results["Targets"].Value); }
    }

    private DataTable ResultsHorizonTargets {
      get { return ((DataTable) Results["Horizon Targets"].Value); }
    }

    private DataTable ResultsStateDetection {
      get { return ((DataTable) Results["State Detection"].Value);}
    }

    private const string ResultsTargetsReal = "Real";
    private const string ResultsTargetRange = "[..]";

    private DataBarSet ResultsSWQualitiesBars {
      get { return (DataBarSet)Results["Ensemble Comparison [SW-Quality]"].Value; }
    }

    //private DataBarSet ResultsSWVotingBars {
    //  get { return (DataBarSet) Results["Ensemble Comparison [SW-Voting]"].Value; }
    //}

    //private DataBarSet ResultsSonarQualitiesBars {
    //  get { return (DataBarSet)Results["Ensemble Comparison [Sonar-Quality]"].Value; }
    //}

    //private DataBarSet ResultsSWPeakQualitiesBars {
    //  get { return (DataBarSet)Results["Ensemble Comparison [SW-PeakQuality]"].Value; }
    //}

    //private double ResultsMeanPR2 {
    //  get { return ((DoubleValue) Results["Mean PR2"].Value).Value; }
    //  set { ((DoubleValue)Results["Mean PR2"].Value).Value = value; }
    //}

    //private double ResultsDetectionAccuracy {
    //  get { return ((DoubleValue)Results["Detection Accuracy"].Value).Value; }
    //  set { ((DoubleValue)Results["Detection Accuracy"].Value).Value = value; }
    //}

    //private double ResultsDetectionExclusivity {
    //  get { return ((DoubleValue)Results["Detection Exclusivity"].Value).Value; }
    //  set { ((DoubleValue)Results["Detection Exclusivity"].Value).Value = value; }
    //}

    protected void SetupResults() {
      //evaluationCount = 0;
      //pR2Sum = 0.0;
      //exclusivitySum = 0.0;
      //accuracySum = 0.0;
      //detectionCount = 0;
      Results.Clear();
      Results.Add(new Result("Sliding Window Movements", new IntValue(0)));
      Results.Add(new Result("Qualities 1", new DataTable("Average Pearson R²")));
      Results.Add(new Result("Qualities 2", new DataTable("Average Horizon Pearson R²")));
      //Results.Add(new Result("Qualities 2", new DataTable("Peak Pearson R²")));
      Results.Add(new Result("ConfidenceTargets", new ExtendedDataTable("ConfidenceTargets")));
      Results.Add(new Result("ErrorRangeTargets", new ExtendedDataTable("ErrorRangeTargets")));
      Results.Add(new Result("Targets", new DataTable("Targets")));
      Results.Add(new Result("Horizon Targets", new DataTable("Horizon Targets")));
      Results.Add(new Result("State Detection", new DataTable("State Detection")));
      Results.Add(new Result("Ensemble Comparison [SW-Quality]", new DataBarSet("Ensemble Comparison [SW-Quality]")));
      //Results.Add(new Result("Ensemble Comparison [SW-Voting]", new DataBarSet("Ensemble Comparison [SW-Voting]")));
      //Results.Add(new Result("Ensemble Comparison [Sonar-Quality]", new DataBarSet("Ensemble Comparison [Sonar-Quality]")));
      //Results.Add(new Result("Ensemble Comparison [SW-PeakQuality]", new DataBarSet("Ensemble Comparison [SW-PeakQuality]")));
      //Results.Add(new Result("Mean PR2", new DoubleValue()));
      //Results.Add(new Result("Detection Accuracy", new DoubleValue()));
      //Results.Add(new Result("Detection Exclusivity", new DoubleValue()));

      ResultsTargets.Rows.Add(new DataRow(ResultsTargetsReal));
      ResultsHorizonTargets.Rows.Add(new DataRow(ResultsTargetsReal));
      ResultsConfidenceTargets.Rows.Add(new RangeDataRow(ResultsTargetsReal));
      ResultsErrorRangeTargets.Rows.Add(new RangeDataRow(ResultsTargetsReal));
      int i = 1;
      foreach (var ensemble in Ensembles) {
        // targets table
        ResultsTargets.Rows.Add(new DataRow(ensemble.Name));
        ResultsHorizonTargets.Rows.Add(new DataRow(ensemble.Name));
        ResultsErrorRangeTargets.Rows.Add(new RangeDataRow(ensemble.Name));
        ResultsErrorRangeTargets.Rows[ensemble.Name].VisualProperties.Color = _colorsBrightPastel[i];
        ResultsConfidenceTargets.Rows.Add(new RangeDataRow(ensemble.Name));
        ResultsConfidenceTargets.Rows[ensemble.Name].VisualProperties.Color = _colorsBrightPastel[i];
        i++;

        ResultsStateDetection.Rows.Add(new DataRow(ensemble.Name));

        // qualities (series)
        //ResultsQualities.Rows.Add(new DataRow(ensemble.Name + " - " + ResultsQualitiesMSE));
        ResultsQualities.Rows.Add(new DataRow(ensemble.Name));
        ResultsHorizonQualities.Rows.Add(new DataRow(ensemble.Name));

        //ResultsPeakQualities.Rows.Add(new DataRow(ensemble.Name));

        // qualities (bars)
        ResultsSWQualitiesBars.Bars.Add(new DataBar(ensemble.Name, ensemble.QualityThreshold.Start, ensemble.QualityThreshold.End));

        // voting (bars)
        //ResultsSWVotingBars.Bars.Add(new DataBar(ensemble.Name, ensemble.ConfidenceThreshold.Start, ensemble.ConfidenceThreshold.End));

        // sonar quality (bars)
        //ResultsSonarQualitiesBars.Bars.Add(new DataBar(ensemble.Name, ensemble.QualityThreshold.Start, ensemble.QualityThreshold.End));

        // quality peaks (bars)
        //ResultsSWPeakQualitiesBars.Bars.Add(new DataBar(ensemble.Name, ensemble.QualityThreshold.Start, ensemble.QualityThreshold.End));
      }

      foreach (var ensemble in Ensembles) {
        Color color = ResultsErrorRangeTargets.Rows[ensemble.Name].VisualProperties.Color;
        color = Color.FromArgb(150, (int)(color.R * 1.0), (int)(color.G * 1.0), (int)(color.B * 1.0));
        string rowName = ensemble.Name + " " + ResultsTargetRange;
        ResultsErrorRangeTargets.Rows.Add(new RangeDataRow(rowName));
        ResultsErrorRangeTargets.Rows[rowName].VisualProperties.Color = color;
        ResultsConfidenceTargets.Rows.Add(new RangeDataRow(rowName));
        ResultsConfidenceTargets.Rows[rowName].VisualProperties.Color = color;
      }
    }

    #endregion

    #region constructors, cloner,...
    public DatastreamAnalysisOptimizer() : base() {
      name = "Datastream Analysis";
      log = new Log();
      results = new ResultCollection();
      ensembles = new ItemList<ProxyEnsembleModel>();
      datastream = new Datastream();
      runsCounter = 0;
      runs = new RunCollection();
      Initialize();
    }

    [StorableConstructor]
    protected DatastreamAnalysisOptimizer(StorableConstructorFlag _) : base(_) {
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      Initialize();
    }

    protected DatastreamAnalysisOptimizer(DatastreamAnalysisOptimizer original, Cloner cloner) : base(original, cloner) {
      name = original.name;
      log = cloner.Clone(original.log);
      results = cloner.Clone(original.results);
      ensembles = (ItemList<ProxyEnsembleModel>) original.Ensembles.Clone(cloner);
      datastream = (Datastream) original.Datastream.Clone(cloner);
      runsCounter = original.runsCounter;
      runs = cloner.Clone(original.runs);
      Initialize();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new DatastreamAnalysisOptimizer(this, cloner);
    }

    private void Initialize() {
      if (runs != null) RegisterRunsEvents();
      if (datastream != null) RegisterDatastreamEvents();
      if (ensembles != null) RegisterEnsembleEvents();
      Prepare();
    }
    #endregion

    #region control actions

    public override void Prepare() {
      if (ensembles == null || ensembles.Count == 0 || datastream == null || !datastream.SlidingWindowEvaluationPossible) return;
      base.Prepare();
      OnPrepared();
    }

    public void Prepare(bool clearRuns) {
      if (ensembles == null || ensembles.Count == 0 || datastream == null || !datastream.SlidingWindowEvaluationPossible) return;

      base.Prepare();
      if (clearRuns) runs.Clear();
      OnPrepared();
    }

    public override void Start(CancellationToken cancellationToken) {
      base.Start(cancellationToken);
      if (ensembles == null || datastream == null) return;

      cancellationTokenSource = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
      stopPending = false;

      if (prepared) {
        SetupResults();
        Datastream.InitializeState();
      }

      Task task = Task.Factory.StartNew(Run, cancellationTokenSource.Token, cancellationTokenSource.Token);
      task.ContinueWith(t => {
        try {
          t.Wait();
        }
        catch (AggregateException ex) {
          try {
            ex.Flatten().Handle(x => x is OperationCanceledException);
          } catch (AggregateException remaining) {
            if(remaining.InnerExceptions.Count == 1) OnExceptionOccurred(remaining.InnerExceptions[0]);
            else OnExceptionOccurred(remaining);
          }
        }
        cancellationTokenSource.Dispose();
        cancellationTokenSource = null;

        // handle stop/pause
        if (stopPending || finished) {
          OnStopped();
        } else {
          OnPaused();
        }
      });
    }

    public override void Pause() {
      if (ensembles == null || datastream == null) return;
      base.Pause();
      cancellationTokenSource.Cancel();
    }

    public override void Stop() {
      if (ensembles == null || datastream == null) return;
      base.Stop();
      if (ExecutionState == ExecutionState.Paused) {
        OnStopped();
      } else {
        stopPending = true;
        cancellationTokenSource.Cancel();
      }
    }

    protected override void OnPrepared() {
      ExecutionTime = TimeSpan.Zero;
      foreach (IStatefulItem statefulItem in this.GetObjectGraphObjects(new HashSet<object>() {Runs}).OfType<IStatefulItem>()) {
        statefulItem.InitializeState();
      }
      results.Clear();
      prepared = true;
      finished = false;
      Log.LogMessage("Datastream analysis prepared");
      base.OnPrepared();
    }

    protected override void OnStarted() {
      Log.LogMessage("Datastream analysis started");
      base.OnStarted();
    }

    protected override void OnPaused() {
      Log.LogMessage("Datastream analysis paused");
      base.OnPaused();
    }

    protected override void OnStopped() {
      try {
        runsCounter++;
        var run = new Run();
        run.Filename = Filename;
        run.Name = string.Format("{0} Run {1}", Name, runsCounter);
        CollectParameterValues(run.Parameters);
        CollectResultValues(run.Results);
        runs.Add(run);
      }
      finally {
        Log.LogMessage("Datastream analysis stopped");
        base.OnStopped();
      }
    }

    #endregion

    #region run
    private void Run(object state) {
      CancellationToken cancellationToken = (CancellationToken) state;
      OnStarted();
      lastUpdateTime = DateTime.UtcNow;
      System.Timers.Timer timer = new System.Timers.Timer(250);
      timer.AutoReset = true;
      timer.Elapsed += new System.Timers.ElapsedEventHandler(timer_Elapsed);
      timer.Start();

      try {
        Run(cancellationToken);
      } catch(Exception ex) {
        if (ex is ArgumentOutOfRangeException) throw ex;
        if (ex is OperationCanceledException) throw ex;
      }
      finally {
        timer.Elapsed -= new System.Timers.ElapsedEventHandler(timer_Elapsed);
        timer.Stop();
        ExecutionTime += DateTime.UtcNow - lastUpdateTime;
      }

      cancellationToken.ThrowIfCancellationRequested();
    }

    private int replayedIndex;

    protected void Run(CancellationToken cancellationToken) {

      if (prepared) {
        replayedIndex = 0;
        prepared = false;


        Setup();

        // play and evaluate initial window
        PlayDatastream();

        for (int i = 1; i < Datastream.EvaluationHorizon.Value; i++) {
          foreach (var ensemble in Ensembles) {
            ResultsHorizonTargets.Rows[ensemble.Name].Values.Add(double.NaN);
          }
        }

        //lastStateValue = Datastream.ProblemData.Dataset.GetDoubleValue("h", replayedIndex);
        if (Datastream.SlidingWindowEvaluationPossible) Evaluate();
        replayedIndex = Datastream.FitnessPartition.End;
      }

      try {
        do {
          while(Datastream.SlidingWindowMovementPossible) {
           cancellationToken.ThrowIfCancellationRequested();

            // perform (delayed) window movement
            Thread.Sleep(Datastream.SlidingWindowMovementDelay.Value);
            Datastream.MoveSlidingWindow();
            ResultsSlidingWindowMovements++;

            // play and evaluate the moved window
            PlayDatastream();
            if (Datastream.SlidingWindowEvaluationPossible) Evaluate();
            replayedIndex = Datastream.FitnessPartition.End;
          }
        } while (Datastream.UpdateAvailable);
        finished = true;
      }
      catch (Exception ex) {
        if (ex is ArgumentOutOfRangeException) throw ex;
        if (ex is OperationCanceledException) throw ex;
      }
      finally {
        // reset everything
        //Prepare(true);
      }
    }

    private void PlayDatastream() {
      var problemData = Datastream.ProblemData;
      var targetVarName = problemData.TargetVariable;

      for (int i = replayedIndex; i < Datastream.FitnessPartition.End; i++) {
        var realValue = problemData.Dataset.GetDoubleValue(targetVarName, i);
        ResultsTargets.Rows[ResultsTargetsReal].Values.Add(realValue);
        ResultsHorizonTargets.Rows[ResultsTargetsReal].Values.Add(realValue);
        ResultsErrorRangeTargets.Rows[ResultsTargetsReal].Values.Add(new Tuple<double, double>(realValue, realValue));
        ResultsConfidenceTargets.Rows[ResultsTargetsReal].Values.Add(new Tuple<double, double>(realValue, realValue));
      }
    }

    //private void Smoothen(double[] estimatedValuesPerRow) {
    //  if (Datastream.SlidingWindowEvaluationScheme.Value == Datastream.EvaluationScheme.equal) {
    //    var avg = estimatedValuesPerRow.Average();
    //    for (int i = 0; i < estimatedValuesPerRow.Length; i++) {
    //      estimatedValuesPerRow[i] = avg;
    //    }
    //  }
    //}

    Dictionary<string, List<double>> estimates;

    private void Setup() {
      estimates = new Dictionary<string, List<double>>();
      foreach(var ensemble in Ensembles) {
        estimates.Add(ensemble.Name, new List<double>());
      }
    }

    private void Evaluate() {

      //evaluationCount++;
      var problemData = Datastream.ProblemData;
      var dataset = problemData.Dataset;
      var targetVarName = problemData.TargetVariable;

      var rows = Enumerable.Range(Datastream.FitnessPartition.Start, Datastream.FitnessPartition.Size);
      var realValues = dataset.GetDoubleValues(targetVarName, rows);

      //int sonarSize = (int)(Datastream.FitnessPartition.Size * Datastream.SlidingWindowSonarRatio.Value);
      //var sonarStart = Datastream.FitnessPartition.End - sonarSize;
      //var sonarRows = Enumerable.Range(sonarStart, sonarSize);
      //var sonarRealValues = dataset.GetDoubleValues(targetVarName, sonarRows);

      var winningEnsembleName = ensembles[0].Name;
      var winningEnsemblePR2 = 0.0;
      var winningEnsembleIdx = 0;
      var ec = 0;
      //bool stateChangeInStep = false;
      var curDetectionCount = 0;
      //var changeDetectionCount = 0;
      //double curStateValue = 0.0;

      foreach (var ensemble in Ensembles) {

        //var sonarEstimatedValuesPerModelPerRow = ensemble.Model.Models.Select(x => x.GetEstimatedValues(datastream.ProblemData.Dataset, sonarRows).ToArray());
        //var sonarEstimatedValuesPerRow = Enumerable.Range(0, sonarSize).Select(r => sonarEstimatedValuesPerModelPerRow.Select(m => m[r]).Average()).ToArray();
        
        var estimatedValuesPerModelPerRow = ensemble.Model.Models.Select(x => x.GetEstimatedValues(datastream.ProblemData.Dataset, rows).ToArray());        
        double[] estimatedValuesPerRow = Enumerable.Range(0, Datastream.FitnessPartition.Size).Select(r => estimatedValuesPerModelPerRow.Select(e => e[r]).Average()).ToArray(); // per row        
        
        //var averageEstimatedValuesPerModel = estimatedValuesPerModelPerRow.Select(x => x.Average()); // per model
        //var averageEstimatedValue = averageEstimatedValuesPerModel.Average();




        // calulate PR
        //var mse = Math.Pow(averageEstimatedValue - realValues.Average(), 2);
        OnlineCalculatorError error;
        var pR = OnlinePearsonsRCalculator.Calculate(estimatedValuesPerRow, realValues, out error);
        var pR2 = error == OnlineCalculatorError.None ? pR * pR : 0.0;

        //var sonarPR = OnlinePearsonsRCalculator.Calculate(sonarEstimatedValuesPerRow, sonarRealValues, out error);
        //var sonarPR2 = error == OnlineCalculatorError.None ? sonarPR * sonarPR : 0.0;
   
        var estimatedValuesPerModelPerRowArr = estimatedValuesPerModelPerRow.ToArray();
        var curPR = 0.0;
        
        //var peakPR = OnlinePearsonsRCalculator.Calculate(estimatedValuesPerModelPerRowArr[0], realValues, out error);
        //for (int i = 1; i < estimatedValuesPerModelPerRow.Count(); i++) {
        //  curPR = OnlinePearsonsRCalculator.Calculate(estimatedValuesPerModelPerRowArr[i], realValues, out error);
        //  if (curPR > peakPR && error == OnlineCalculatorError.None) {
        //    peakPR = curPR;
        //  }
        //}
        //var peakPR2 = peakPR * peakPR;

        // calculate
        double mae = OnlineMeanAbsoluteErrorCalculator.Calculate(realValues, estimatedValuesPerRow, out error);
        //double mse = OnlineMeanSquaredErrorCalculator.Calculate(realValues, estimatedValuesPerRow, out error);

        // calculate confidence
        //var cm1 = CalulateEnsembleConfidenceCM1(ensemble, estimatedValuesPerModelPerRow, realValues);

        int replayAmount = Datastream.FitnessPartition.End - replayedIndex;
        int replayCount = estimatedValuesPerRow.Length - replayAmount;

        //Smoothen(estimatedValuesPerRow);

        for (int i = replayedIndex; i < Datastream.FitnessPartition.End; i++) {
          //ResultsTargets.Rows[ensemble.Name].Values.Add(averageEstimatedValue);
          
          var estVal = estimatedValuesPerRow[replayCount];
          var estStdDev = ComputeStdDev(estimatedValuesPerModelPerRow.Select(x => x[replayCount]).ToList());

          ResultsTargets.Rows[ensemble.Name].Values.Add(estimatedValuesPerRow[replayCount]);
          if(Datastream.RollingEvaluationHorizon.Value) {
            var rollingPrediction = RollingEvaluate(ensemble, i);
            ResultsHorizonTargets.Rows[ensemble.Name].Values.Add(rollingPrediction);
          } else {
            ResultsHorizonTargets.Rows[ensemble.Name].Values.Add(estimatedValuesPerRow[replayCount]);
          }


          ResultsErrorRangeTargets.Rows[ensemble.Name].Values.Add(new Tuple<double, double>(estVal, estVal));
          ResultsErrorRangeTargets.Rows[ensemble.Name + " " + ResultsTargetRange].Values.Add(new Tuple<double, double>(estVal - mae, estVal + mae));
          ResultsConfidenceTargets.Rows[ensemble.Name].Values.Add(new Tuple<double, double>(estVal, estVal));
          ResultsConfidenceTargets.Rows[ensemble.Name + " " + ResultsTargetRange].Values.Add(new Tuple<double, double>(estVal - 1.96*estStdDev, estVal + 1.96*estStdDev));
          replayCount++;

          //ResultsQualities.Rows[ensemble.Name + " - " + ResultsQualitiesMSE].Values.Add(mse);
          ResultsQualities.Rows[ensemble.Name].Values.Add(pR2);
          //ResultsPeakQualities.Rows[ensemble.Name].Values.Add(peakPR2);

          double amp = (pR2 > ensemble.QualityThreshold.Start) ? 1.0 : 0.0;
          ResultsStateDetection.Rows[ensemble.Name].Values.Add(amp);    
        }
        ResultsSWQualitiesBars.Bars[ensemble.Name].Value = pR2;
        //ResultsSonarQualitiesBars.Bars[ensemble.Name].Value = sonarPR2;
        //ResultsSWVotingBars.Bars[ensemble.Name].Value = cm1;
        //ResultsSWPeakQualitiesBars.Bars[ensemble.Name].Value = peakPR2;

        if (pR2 > ensemble.QualityThreshold.Start)
          curDetectionCount++;

        if (pR2 > winningEnsemblePR2) {
          winningEnsemblePR2 = pR2;
          winningEnsembleName = ensemble.Name;
          winningEnsembleIdx = ec;
        }
        ec++;
      }

      // quality meassures
      // =================================================

      //if (winningEnsemblePR2 > ensembles[winningEnsembleIdx].QualityThreshold.Start) {
      //  // Mean PR2
      //  pR2Sum += winningEnsemblePR2;
      //  detectionCount++;
      //  ResultsMeanPR2 = pR2Sum/detectionCount;

      //  // detection exclusivity --> 1.0 - (x - 1) * (1.0 / n)
      //  var fraction = 1.0 / ensembles.Count;
      //  var curExclusivity = 1.0 - (curDetectionCount - 1) * fraction;
      //  exclusivitySum += curExclusivity;
      //  ResultsDetectionExclusivity = exclusivitySum / detectionCount;
      //}

      OnlineCalculatorError err;

      int size = Datastream.FitnessPartition.Size;
      int horizon = Datastream.EvaluationHorizon.Value;
      int start = replayedIndex - size;
      int end = start + size;

      var realVals = ResultsHorizonTargets.Rows[ResultsTargetsReal].Values;
      if (replayedIndex >= (size - 1)+horizon && realVals.Count >= end) {
        
        var reals = Enumerable.Range(start, size).Select(r => realVals[r]).ToArray();

        foreach (var ensemble in Ensembles) {
          var estimatedVals = ResultsHorizonTargets.Rows[ensemble.Name].Values;

          if (estimatedVals.Count >= end) {
            var vals = Enumerable.Range(start, size).Select(r => estimatedVals[r]).ToArray();
            var pR = OnlinePearsonsRCalculator.Calculate(vals, reals, out err);
            var pR2 = err == OnlineCalculatorError.None ? pR * pR : 0.0;
            ResultsHorizonQualities.Rows[ensemble.Name].Values.Add(pR2);
          }
        }
      }


    }

    private double RollingEvaluate(ProxyEnsembleModel ensemble, int index) {
      
      var variables = datastream.ProblemData.InputVariables
        .Select(x => x.Value)
        .Where(x => x != datastream.ProblemData.TargetVariable)
        .ToList();
      var ods = datastream.ProblemData.Dataset;
      var values = new double[1,variables.Count];
      var rows = Enumerable.Range(0, 1);

      for(int i = 0; i < variables.Count; i++) {
        values[0,i] = ods.GetDoubleValue(variables[i], index);
      }

      Dataset ds = new Dataset(variables, values);
      double estimate = ensemble.Model.GetEstimatedValues(ds, rows).ToArray()[0];

      for(int i = 1; i < datastream.EvaluationHorizon.Value; i++) {
        values[0, 0] = estimate;
        for(int v = 0; v < variables.Count-1; v++) {
          values[0, v + 1] = ds.GetDoubleValue(variables[v], 0);
        }
        ds = new Dataset(variables, values);
        estimate = ensemble.Model.GetEstimatedValues(ds, rows).ToArray()[0];
      }
      
      return estimate;
    }

    private bool accDiscount = false;

    private double CalulateEnsembleConfidenceCM1(ProxyEnsembleModel ensemble, IEnumerable<double[]> estimatedValuesPerModelPerRow, IEnumerable<double> realValues) {
      int noModels = estimatedValuesPerModelPerRow.Count();
      int votes = 0;

      OnlineCalculatorError error;

      foreach(var model in estimatedValuesPerModelPerRow) {
        var pR = OnlinePearsonsRCalculator.Calculate(model, realValues, out error);
        var pR2 = error == OnlineCalculatorError.None ? pR * pR : 0.0;
        if (pR2 >= ensemble.QualityThreshold.End) {
          votes++;
        }
      }
      

      return (double)votes/noModels;
    }

    private void timer_Elapsed(object sender, System.Timers.ElapsedEventArgs e) {
      System.Timers.Timer timer = (System.Timers.Timer)sender;
      timer.Enabled = false;
      DateTime now = DateTime.UtcNow;
      ExecutionTime += now - lastUpdateTime;
      lastUpdateTime = now;
      timer.Enabled = true;
    }

    public void CollectParameterValues(IDictionary<string, IItem> values) {
      values.Add("Datastream Analysis Name", new StringValue(Name));
      if (Datastream != null) {
        Datastream.CollectParameterValues(values);
        values.Add("Datastream Name", new StringValue(Datastream.Name));
      }
    }

    public void CollectResultValues(IDictionary<string, IItem> values) {
      values.Add("Execution Time", new TimeSpanValue(ExecutionTime));
      Results.CollectResultValues(values);
    }
    #endregion

    #region events

    #region events registration

    public EventHandler EnsemblesChanged;
    public EventHandler DatastreamChanged;

    protected virtual void DeregisterRunsEvents() {
      runs.CollectionReset -= new CollectionItemsChangedEventHandler<IRun>(Runs_CollectionReset);
    }

    protected virtual void RegisterRunsEvents() {
      runs.CollectionReset += new CollectionItemsChangedEventHandler<IRun>(Runs_CollectionReset);
    }

    protected virtual void RegisterDatastreamEvents() {
      datastream.Reset += new EventHandler(Datastream_Reset);
      datastream.ProblemDataChanged += new EventHandler(Datastream_ProblemDataChanged);
    }

    protected virtual void DeregisterDatastreamEvents() {
      datastream.Reset -= new EventHandler(Datastream_Reset);
      datastream.ProblemDataChanged -= new EventHandler(Datastream_ProblemDataChanged);
    }

    protected virtual void RegisterEnsembleEvents() {
      ensembles.ItemsAdded += new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_ItemsChanged);
      ensembles.ItemsMoved += new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_ItemsChanged);
      ensembles.ItemsRemoved += new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_ItemsChanged);
      ensembles.ItemsReplaced += new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_ItemsChanged);
      ensembles.CollectionReset += new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_Reset);
    }

    protected virtual void DeregisterEnsembleEvents() {
      ensembles.ItemsAdded -= new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_ItemsChanged);
      ensembles.ItemsMoved -= new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_ItemsChanged);
      ensembles.ItemsRemoved -= new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_ItemsChanged);
      ensembles.ItemsReplaced -= new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_ItemsChanged);
      ensembles.CollectionReset -= new CollectionItemsChangedEventHandler<IndexedItem<ProxyEnsembleModel>>(Ensembles_Reset);
    }
    #endregion

    #region event handling

    protected virtual void Runs_CollectionReset(object sender, CollectionItemsChangedEventArgs<IRun> e) {
      runsCounter = runs.Count;
    }

    protected virtual void Datastream_Reset(object sender, EventArgs e) {
      Prepare();
    }

    protected virtual void Datastream_ProblemDataChanged(object sender, EventArgs e) {
      Prepare();
    }

    protected virtual void Ensembles_Reset(object sender, EventArgs e) {
      Prepare();
    }

    protected virtual void Ensembles_ItemsChanged(object sender, EventArgs e) {
      Prepare();
    }

    private void OnEnsemblesChanged() {
      var changed = EnsemblesChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    private void OnDatastreamChanged() {
      var changed = DatastreamChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    #endregion

    #endregion

    #region NamedItem

    [Storable]
    protected string name;

    /// <inheritdoc/>
    /// <remarks>Calls <see cref="OnNameChanging"/> and also <see cref="OnNameChanged"/> 
    /// eventually in the setter.</remarks>
    public string Name {
      get { return name; }
      set {
        if (!CanChangeName) throw new NotSupportedException("Name cannot be changed.");
        if (!(name.Equals(value) || (value == null) && (name == string.Empty))) {
          CancelEventArgs<string> e = value == null
            ? new CancelEventArgs<string>(string.Empty)
            : new CancelEventArgs<string>(value);
          OnNameChanging(e);
          if (!e.Cancel) {
            name = value == null ? string.Empty : value;
            OnNameChanged();
          }
        }
      }
    }

    public virtual bool CanChangeName {
      get { return true; }
    }

    [Storable] protected string description;

    public string Description {
      get { return description; }
      set {
        if (!CanChangeDescription) throw new NotSupportedException("Description cannot be changed.");
        if (!(description.Equals(value) || (value == null) && (description == string.Empty))) {
          description = value == null ? string.Empty : value;
          OnDescriptionChanged();
        }
      }
    }

    public virtual bool CanChangeDescription {
      get { return true; }
    }

    /// <summary>
    /// Gets the string representation of the current instance in the format: <c>Name: [null|Value]</c>.
    /// </summary>
    /// <returns>The current instance as a string.</returns>
    public override string ToString() {
      return Name;
    }

    /// <inheritdoc/>
    public event EventHandler<CancelEventArgs<string>> NameChanging;

    /// <summary>
    /// Fires a new <c>NameChanging</c> event.
    /// </summary>
    /// <param name="e">The event arguments of the changing.</param>
    protected virtual void OnNameChanging(CancelEventArgs<string> e) {
      var handler = NameChanging;
      if (handler != null) handler(this, e);
    }

    /// <inheritdoc/>
    public event EventHandler NameChanged;

    /// <summary>
    /// Fires a new <c>NameChanged</c> event.
    /// </summary>
    /// <remarks>Calls <see cref="ItemBase.OnChanged"/>.</remarks>
    protected virtual void OnNameChanged() {
      var handler = NameChanged;
      if (handler != null) handler(this, EventArgs.Empty);
      OnToStringChanged();
    }

    /// <inheritdoc/>
    public event EventHandler DescriptionChanged;

    /// <summary>
    /// Fires a new <c>DescriptionChanged</c> event.
    /// </summary>
    /// <remarks>Calls <see cref="ItemBase.OnChanged"/>.</remarks>
    protected virtual void OnDescriptionChanged() {
      var handler = DescriptionChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    #endregion nameditem

    #region Helpers
    private double ComputeStdDev(List<double> series) {
      double avg = series.Average();
      double sumOfSquaredDeviations = series.Sum(val => (val - avg) * (val - avg));
      return Math.Sqrt(sumOfSquaredDeviations / (series.Count - 1));
    }
    #endregion
  }
}